
import processing.serial.*;
import gab.opencv.*;
import java.awt.Rectangle;

int GSType = 1;
boolean inversePolar = false;
boolean isPortrait = true;
String[] boardName = {
  "32x16", "32x16", "32x24H", "32x24H", "32x24H", "42x32H", "42x24H", "42x24H"
};
int[] intplFactor  = {
  1, 1, 2, 2, 1, 2, 2, 1
};
int[] boardAmnt    = {
  1, 2, 1, 2, 2, 1, 2, 2
};
int [] boardR      = {
  32, 32, 32, 32, 32, 42, 42, 42
};
int [] boardC      = {
  16, 16, 12, 12, 12, 16, 16, 16
};

//Serial
Serial port = null;
Serial port2 = null;

//Const
int THLD = 5; 
int CONST_MAX  = 20;
float MAX_SENS = CONST_MAX;
float SENS     = MAX_SENS-5;
float low_thr = 30;
float high_thr = 60;
float x_interval = 1.2f;//3.766f/3;
float y_interval = 1.2f;//3.774f/3;
float x_padding = 5.0f;
float y_padding = 5.0f;
int info_padding = 16;
int NOISE_THRESHOLD = 20;

//SensorData
int boardNum = boardAmnt[GSType];
int singleR  = boardR[GSType];
int singleC  = boardC[GSType];
int totalR   = singleR;
int totalC   = singleC*boardNum*intplFactor[GSType];
float[][] initSensorVal = new float[totalR][totalC];
float[][] hallSensorVal = new float[totalR][totalC];

//SampleData
BicubicInterpolator   bcp = new BicubicInterpolator();
BiLinearInterpolator  blp = new BiLinearInterpolator();
int sampleRes = 10;
int sampleR   = totalR * (sampleRes);
int sampleC   = totalC * (sampleRes);
float[][] sampleMVal  = new float[sampleR][sampleC];


//Stylus
float maxM = -1; //M = magnitude
float minM = 1;

boolean initialize1 = true;
boolean initialize2 = true;
boolean flipSide = false;
boolean drawSensors = true;
boolean drawFPS = false;
boolean drawTilt = true;
boolean approximate = false;
boolean showRaw = false;
boolean drawDT = false;

boolean drawPair = true;
boolean dumpData = true;
boolean drawLayerDetail = false;



PImage rawMap, sensors;
OpenCV opencv;
ArrayList<PVector> L1centroids;

//Stone Only
int CALIB_THLD = 50;
boolean initCalib = false;

int SERIES_NUM = 6;
int SERIES_SPACE = 17;
int SERIES_BASE = 18;
PImage[] MapN = new PImage[SERIES_NUM];
PImage[] MapS = new PImage[SERIES_NUM];
ArrayList<ArrayList<Contour>> contoursNList = new ArrayList<ArrayList<Contour>>(SERIES_NUM);
ArrayList<ArrayList<Contour>> contoursSList = new ArrayList<ArrayList<Contour>>(SERIES_NUM);
ArrayList<GData> gData;
//

int scrImgWidth = int(x_padding*2+(float)((totalC-1)*sampleRes*x_interval));
int scrImgHeight = int(y_padding*2+(float)((totalR-1)*sampleRes*y_interval));
float imgWidth = (float)((totalC-1)*sampleRes*x_interval);
float imgHeight = (float)((totalR-1)*sampleRes*y_interval);
boolean doCalibration = true;
int t_end = millis();

String calibFileName = GSType+".txt";

void setup() {
  frameRate(200);
  
  if (isPortrait) {
    size(scrImgWidth, scrImgHeight+info_padding*3+(int)y_padding);
  } else {
    size(scrImgHeight, scrImgWidth+info_padding*2+(int)y_padding);
  }
  
  textSize(12);
  stroke(0);
  ellipseMode(CENTER);

  // List all the available serial ports
  String[] serialList = Serial.list();
  println("Found " + serialList.length + " serial device(s).");
  if (serialList.length == 0)
    exit();
  for (int i = 0; i < serialList.length; i++)
    println("["+(i+1)+"] "+serialList[i]);
  if (boardNum > 1 && serialList.length > 1) {
    port2 = new Serial(this, serialList[serialList.length-1], 921600);  // make sure Arduino is talking serial at this baud rate
    port2.write("a");
  }
  port = new Serial(this, serialList[serialList.length-boardNum], 921600);  // make sure Arduino is talking serial at this baud rate
  port.write("a");
  
  if (isPortrait)
    rawMap = createImage((totalC-1)*sampleRes, (totalR-1)*sampleRes, RGB);
  else
    rawMap = createImage((totalR-1)*sampleRes, (totalC-1)*sampleRes, RGB);
  
  opencv = new OpenCV(this, rawMap);
//  L1centroids = new ArrayList<PVector>();
  gData = new ArrayList<GData>();

  background(0);
  fill(76);
  noStroke();
  if (isPortrait) {
    for (int i = 0 ; i < totalC*totalR; i++)
      ellipse(((int)((totalC-1)-i%totalC)*sampleRes)*x_interval+x_padding, ((int)(i/totalC)*sampleRes)*y_interval+y_padding, x_interval, y_interval);
  }
  else {
    for (int i = 0 ; i < totalC*totalR; i++)
      ellipse(((int)(i/totalC)*sampleRes)*x_interval+x_padding, ((int)((totalC-1)-i%totalC)*sampleRes)*y_interval+y_padding, x_interval, y_interval);
  }
  saveFrame("sensors.jpg");
  sensors = loadImage("sensors.jpg");
  t_end = millis();
  
  File file = new File(sketchPath(calibFileName));
  if (!file.exists()) {
    doCalibration = true;
    println("First-Time Calibration");
  }
  else {
    initSensorVal = loadCalibFile(calibFileName);
    doCalibration = false;
    initCalib = true;  
    println("Found it");
  }
}

void draw() {
  if (doCalibration) {
    initialize1 = false;
    initialize2 = false;
    doCalibration = false;
  } 
  else {
    background(0);
    upSampling();
    updateClipping();
    if (showRaw) drawRawData();
    else image(sensors, 0, 0);
    drawTokens();
    drawInfo();
    
//    if(dumpData) dumpTokens();
  }
  t_end = millis();
  // Receive message from serial port
  getSerialMsg();
}

void keyReleased() {
  if (key == 'f' || key == 'F') {
    doCalibration = true;
    flipSide = !flipSide;
  } 
  else if (key == 'r' || key == 'R')
    showRaw = !showRaw;
  else if (key == 'c' || key == 'C') {
    doCalibration = true;
  }
  else if (key == 'i' || key == 'I'){
    ++SERIES_BASE;
  }
  else if (key == 'k' || key == 'K'){
    --SERIES_BASE;
  }
  else if (key == 'j' || key == 'J'){
    ++SERIES_SPACE;
  }
  else if (key == 'l' || key == 'L'){
    --SERIES_SPACE;
  }
  else if (key == '1'){
    SERIES_BASE = 32;
  }
  else if (key == '2'){
    SERIES_BASE = 18;
  }
  
}
